#!/bin/bash

#---------------------- CONFIGURATION -------------------------
#Use LAM MPI (0 - false(default), 1-true)
LAM="0"

#Optput file (it will be in either in 'bin32' or in 'bin64' folder)
OUTPUT="Tesseral2D.lexe"

#MPI compiler for C++ 
#MPICC="mpiCC"

#Enable GPU (CUDA/OpenCL)
#GPU="1"

#---------------------------------------------------------------

ask_yes_no ()
{
    unset ANSWER
    local PROMPT="$1"
    shift
    local DEFAULT_ANSWER="$1"
    while [[ "$ANSWER" != [YyNn] ]]; do
	echo -n "$PROMPT"
	read ANSWER
	if [[ -z "$ANSWER" ]]; then
	    echo "Selected answer = $DEFAULT_ANSWER"
	    ANSWER=$DEFAULT_ANSWER
	fi
	if [[ "$ANSWER" != [YyNn] ]]; then
	    #clear
	    echo 'Your selection is not vaild, please try again.'
	    echo
	fi
    done
    # Set to 0/1
    if [[ "$ANSWER" =~ [Yy] ]]; then
	ANSWER=1
    else
	ANSWER=0
    fi
}

ask_path_file ()
{
    unset ANSWER
    local PROMPT="$1"
    shift
    local DEFAULT_ANSWER="$1"

    while [[ ! -x "$ANSWER" ]]; do
        echo -n "$PROMPT"
	read ANSWER
        if [[ -z "$ANSWER" ]]; then
	    echo "Selected answer = $DEFAULT_ANSWER"
	    ANSWER=$DEFAULT_ANSWER
	fi
        if [[ ! -x "$ANSWER" ]]; then
	    #clear
	    echo "Unable to execute $ANSWER, please try again."
	    echo
        fi
    done
}

echo
echo " Build 2D modeling engine for Linux Cluster using custom MPI implementation"
echo " (C) Tesseral Technologies Inc."
echo 

#Optimization options
OPT="-O3 -Wl,-rpath,."

NO_COLOR='[0m'
GREEN='[0;32m'
RED='[0;31m'
YELLOW='[0;33m'

MPIFOUND="0"

if [ "$MPICC" = "" ]; then
    if [ -f /usr/lib64/openmpi/bin/mpic++ ]; then
        MPICC="/usr/lib64/openmpi/bin/mpic++"
        LAM="0"
    else
        MPICC="mpiCC"
    fi
fi

which $MPICC &>/dev/null
if [ $? -eq 0 ]; then
    MPIFOUND="1"
else
    MPICC="mpicxx"
    which $MPICC &>/dev/null
    if [ $? -eq 0 ]; then
        MPIFOUND="1"
    else
        MPICC="mpic++"
        which $MPICC &>/dev/null
        if [ $? -eq 0 ]; then
            MPIFOUND="1"
        fi
    fi
fi

if [ $MPIFOUND -eq 0 ]; then
    ask_path_file "Please, specify MPI C++ compiler pathname (e.g. /bin/mpic++) [mpic++]:" "mpic++"
    MPICC="$ANSWER"
    which $MPICC &>/dev/null
    if [ $? -eq 0 ]; then
        MPIFOUND="1"
    fi
fi

if [ $MPIFOUND -eq 0 ]; then
    echo "$RED MPI C++ compiler not found! (Please, set MPICC environment variable) $NO_COLOR"
    exit -1
fi

if [ -z $GPU ]; then
    ask_yes_no "Enable GPU support [Y/n]?" "y"
    GPU=$ANSWER
fi

if [ $LAM -eq 0 ]; then
    LAM_TEST=`$MPICC --showme 2>/dev/null | grep "\-llam"`
    if [ -n "$LAMTEST" ]; then
	LAM="1"
    fi
fi 

DLIB="-L ./lib"
TLIB="-lTesseral2D64g412"
LIBS="-lm -lpthread -ldl"
OPTLAM=""

if [ $LAM -ne 0 ]; then
   OPTLAM="-D_LAM_DEFINE"
fi

MACHINE=`uname -m`
GCCVER=`gcc -dumpversion`

GCC_SUPPORT="0"

echo "OS Architecture: $YELLOW $MACHINE $NO_COLOR"
echo "GCC version: $YELLOW $GCCVER $NO_COLOR"

TLIB=""
if [ "$MACHINE" = "x86_64" ]; then
    OUTDIR="bin64/"
    LIBS="$LIBS -ltess_common64"
    TLIB="-lTesseral2D64g346"
    if [ $GPU -ne 0 ]; then
        LIBS="$LIBS -lcudart64 -lOpenCL-64"
    fi
else
    OUTDIR="bin32/"
    LIBS="$LIBS"
    TLIB="-lTesseral2D32g346"
    if [ $GPU -ne 0 ]; then
        LIBS="$LIBS -lcudart32 -lOpenCL-32"
    fi
fi

if [[ "$GCCVER" > "6" ]]; then
   TLIB="-no-pie $TLIB"
fi

if [ "$GCCVER" = "3.4.6" ]; then
    GCC_SUPPORT="1"
    if [ "$MACHINE" = "x86_64" ]; then
        TLIB="-lTesseral2D64g346"
    else
        TLIB="-lTesseral2D32g346"
    fi
fi

if [ "$GCCVER" = "4.1.2" ]; then
    GCC_SUPPORT="1"
    if [ "$MACHINE" = "x86_64" ]; then
        TLIB="-lTesseral2D64g412"
    else
        TLIB="-lTesseral2D32g412"
    fi
fi

if [[ "$GCCVER" =~ 4\.4\.[0-9] ]]; then
    GCC_SUPPORT="1"
    if [ "$MACHINE" = "x86_64" ]; then
        TLIB="-lTesseral2D64g446"
    else
        TLIB="-lTesseral2D32g446"
    fi
fi

if [[ "$GCCVER" =~ 4\.8\.[0-9] ]]; then
    GCC_SUPPORT="1"
    TLIB="-lTesseral2D64g485"
fi

if [[ "$GCCVER" =~ 7\.[0-9]\.[0-9] ]]; then
    GCC_SUPPORT="1"
    TLIB="-lTesseral2D64g740"
fi

if [ "$GCC_SUPPORT" = "0" ]; then
    # Find closest gcc match (x64 only)
    if [ "$MACHINE" = "x86_64" ]; then
	if [[ "$GCCVER" =~ 3\.[0-9]\.[0-9] ]]; then
	    TLIB="-lTesseral2D64g346"
	fi
	if [[ "$GCCVER" =~ 4\.[0-9]\.[0-9] ]]; then
	    TLIB="-lTesseral2D64g446"
	fi
    fi
fi

if [ $GPU -ne 0 ]; then
    TLIB="${TLIB}c"
fi

if [ $GCC_SUPPORT -eq 0 ]; then
    echo "$YELLOW Warning GCC version $GCCVER is not officially supported. $NO_COLOR"
fi

$MPICC src/main.cpp $OPT $OPTLAM -o $OUTDIR$OUTPUT $DLIB $TLIB $LIBS $@ &> errors.txt

if [ $? -eq 0 ]; then
    echo "$GREEN Compilation has been SUCCEEDED! $NO_COLOR ($OUTDIR$OUTPUT)"
    rm -f errors.txt
else
    echo "Command used to build the binary:" >> errors.txt
    echo "$MPICC src/main.cpp $OPT $OPTLAM -o $OUTDIR$OUTPUT $DLIB $TLIB $LIBS $@" >> errors.txt
    echo "$RED Compilation has been FAILED! (Please, sent errors.txt to developers) $NO_COLOR"
    exit -1
fi
