// Mod25D.cpp : Defines the entry point for the console application.
//

#include "StdAfx_exe.h"

#include <unistd.h>

int g_nThreadsPerNode=2;
int g_nThreadsPerProcess=2;

//#define DEBUG_MPI_STARTUP

#ifdef _UNIX_

#ifdef _LAM_DEFINE
#undef SEEK_CUR 
#undef SEEK_END
#undef SEEK_SET
#endif

#include <time.h>
#include <stdlib.h>
static void msleep2(unsigned int milliseconds)
{
	struct timespec requested_time, remaining;

	requested_time.tv_sec=(int)(milliseconds/1000);
	requested_time.tv_nsec=(milliseconds % 1000) * 1000000;

	nanosleep(&requested_time, &remaining);
}

#endif //_UNIX_

static int GetNumCores()
{
	int nCores = 1;
	nCores = sysconf( _SC_NPROCESSORS_ONLN );
	return nCores;
}

#ifdef CLUSTER_LAM_MPI
#include <mpi.h>

#include "MpiFunctions.h"

extern int g_nNode;
extern int g_nSize;

MPI_Group MPI_GROUP_WORLD, MPI_GROUP_MT; 
MPI_Comm MPI_COMM_MT; 
#define TAG_HOSTNAME	20
#define TAG_DIE			21

struct _t_node_rank
{
	char szName[80];
	int rank;
	bool bSelected;
};

extern "C"
{
int MPI_Comm_size_our()
{
	int nSize;
	MPI_Comm_size(MPI_COMM_MT, &nSize);
	return nSize;
}

int MPI_Comm_rank_our()
{
	int nNode;
	MPI_Comm_rank(MPI_COMM_MT, &nNode);
	return nNode;
}

void MPI_Finalize_our()
{
	MPI_Finalize();
}

void MPI_Send_char(char* str, int len, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Send (str,len,MPI_CHAR,nNode,nType,MPI_COMM_MT);
}

void MPI_Send_int(int* n, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Send (n,cnt,MPI_INT,nNode,nType,MPI_COMM_MT);
}


void MPI_Send_float(float* f, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Send (f,cnt,MPI_FLOAT,nNode,nType,MPI_COMM_MT);
}

void MPI_Send_long(long* l, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Send (l,cnt,MPI_LONG,nNode,nType,MPI_COMM_MT);
}

int MPI_Recv_char(char* str, int len, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Status status;
	MPI_Recv (str,len,MPI_CHAR,nNode,nType,MPI_COMM_MT,&status);
	return status.MPI_SOURCE;
}

int MPI_Recv_int(int* n, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Status status;
	MPI_Recv (n,cnt,MPI_INT,nNode,nType,MPI_COMM_MT,&status);
	return status.MPI_SOURCE;
}

int MPI_Recv_float(float* f, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Status status;
	MPI_Recv (f,cnt,MPI_FLOAT,nNode,nType,MPI_COMM_MT,&status);
	return status.MPI_SOURCE;
}

int MPI_Recv_long(long* l, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Status status;
	MPI_Recv (l,cnt,MPI_LONG,nNode,nType,MPI_COMM_MT,&status);
	return status.MPI_SOURCE;
}

void MPI_Barrier_our()
{
	MPI_Barrier(MPI_COMM_MT);
}

void MPI_Bcast_int(int* n, int cnt, int nNode)
{
	MPI_Bcast(n,cnt,MPI_INT,nNode,MPI_COMM_MT);
}

int MPI_Recv_double(double* f, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Status status;
	MPI_Recv (f,cnt,MPI_DOUBLE,nNode,nType,MPI_COMM_MT,&status);
	return status.MPI_SOURCE;
}

void MPI_Send_double(double* f, int cnt, int nNode, int nType)
{
	if(nNode==MPI_ANY_SOURCE_OUR)
		nNode = MPI_ANY_SOURCE;
	MPI_Send (f,cnt,MPI_DOUBLE,nNode,nType,MPI_COMM_MT);
}

}
#endif //CLUSTER_LAM_MPI


int mod2d_run(int nFirstShot, int nLastShot);
extern char g_szExeDir[_MAX_PATH];
extern char g_szProjDir[_MAX_PATH];

int main(int argc, char* argv[])
{
	g_szExeDir[0] = '\0';

	int pid = getpid();
	char proc_exe[_MAX_PATH];
	sprintf(proc_exe, "/proc/%d/exe", pid);
	int nr = readlink(proc_exe, g_szExeDir, _MAX_PATH);
	g_szExeDir[nr]=0;
	char* pS = strrchr(g_szExeDir,U_SLASH);
	if(pS!=NULL)
	{
	    pS[0+1]='\0';
	}

	int nFirstShot=0, nLastShot=0;
	if(argc>=4)
	{
		nFirstShot=atoi(argv[2]); nLastShot=atoi(argv[3]);
	}

#ifdef CLUSTER_LAM_MPI
	MPI_Init(&argc, &argv);
	int nRank=0, nSize=0;
	int i,j;
	_t_node_rank *nodes = NULL;
	MPI_Status status;

	MPI_Comm_rank(MPI_COMM_WORLD, &nRank);
	MPI_Comm_size(MPI_COMM_WORLD, &nSize);

	if (nSize==0) return -2;

	int nUniqueCount=0;
	char szHostName[80];
	int *members = new int[nSize+1];
	memset(members, 0, (nSize+1)*sizeof(int));

	if(gethostname(szHostName,80-1)!=0)
		strcpy(szHostName,"");

#ifndef _MULTITHREADED
	g_nThreadsPerNode=1;
#else //_MULTITHREADED
	g_nThreadsPerNode=GetNumCores();
	g_nThreadsPerProcess=g_nThreadsPerNode;
	//Reading number of cores in Runtask.ini
	FILE *fini = fopen("runtask.ini","r");
	if (fini!=NULL)
	{
		char buf[255];
		char val[128];
		bool bOpt = false;
		while (!feof(fini))
		{
			fgets(buf,254,fini);

			if(strncmp(buf,"[optimization]",14)==0)
				bOpt=true;

			if(strncmp(buf,"threads per process",19))
				continue;

			if(strncmp(buf,"threads per process",19) || !bOpt)
				continue;

			char *tok = (char *)strchr(buf+19,'=');
			if(tok==NULL) continue;
			tok++;
			while(*tok!=0 && (BYTE)*tok<=(BYTE)' ') tok++;
			int n=__min((int)strlen(tok),127);
			if(n>0)
			{
				strncpy(val,tok,n); val[n]=0;
			}
			else val[0]=0;
			int cnt = atoi(val);
			if (cnt>0 && cnt<g_nThreadsPerProcess)
				g_nThreadsPerProcess=cnt;
		}
		fclose(fini);
	}
#endif //_MULTITHREADED
	if (g_nThreadsPerNode==1 || nSize==1)
	{
		MPI_COMM_MT = MPI_COMM_WORLD;
		g_nNode = nRank; 
		g_nSize = nSize; 
	}
	else
	{
		int nTicketsPerNode = (int) (g_nThreadsPerNode/g_nThreadsPerProcess);
		if (nTicketsPerNode<1) nTicketsPerNode=1;
		//Determining unique nodes
		if (nRank==0)
		{
#ifdef DEBUG_MPI_STARTUP
			printf("...Assuming %d threads per process, %d threads per node, %d processes per node\n", 
				g_nThreadsPerProcess, g_nThreadsPerNode, nTicketsPerNode); fflush(stdout);
#endif //DEBUG_MPI_STARTUP
			//Recieving node names
			char szHost[80];

			nodes = new _t_node_rank[nSize+1];
			memset(nodes, 0, (nSize+1)*sizeof(_t_node_rank) );
			nodes[0].bSelected = true;
			for (i=0; i<nSize-1; i++)
			{				
				MPI_Recv (szHost,80,MPI_CHAR,MPI_ANY_SOURCE,TAG_HOSTNAME,MPI_COMM_WORLD,&status);
				nodes[status.MPI_SOURCE].rank = status.MPI_SOURCE;
				nodes[status.MPI_SOURCE].bSelected = true;
				strncpy(nodes[status.MPI_SOURCE].szName, szHost, 80);			
			}

			//Selecting unique nodes;
			for (i=1; i<nSize; i++)
			{
				int ncnt = nTicketsPerNode;
				if (nodes[i].bSelected==true)
				{
					strcpy(szHost, nodes[i].szName);
					for (j=1; j<nSize; j++)
					{
						if (strcmp(szHost, nodes[j].szName)==0)
						{
							ncnt--;
							if (ncnt<0)	nodes[j].bSelected=false;
						}							
					}
				}
			}
			nUniqueCount=1;
			members[0]=0;
			for (i=1; i<nSize; i++)
			{
#ifdef DEBUG_MPI_STARTUP
				if (nodes[i].bSelected)
					printf("--Node: rank %d, %s, Active\n", i, nodes[i].szName);
				else
					printf("--Node: rank %d, %s, Passive\n", i, nodes[i].szName);
#endif //DEBUG_MPI_STARTUP
				fflush(stdout);
				if (nodes[i].bSelected)
				{
					members[nUniqueCount]=i;
					nUniqueCount++;
				}
			}
		}
		else
		{
			int len = strlen(szHostName);
			MPI_Send (szHostName,80, MPI_CHAR,0,TAG_HOSTNAME,MPI_COMM_WORLD);		
		}
		//Broadcasting members
		MPI_Bcast (&nUniqueCount,1, MPI_INT,0,MPI_COMM_WORLD);		
		MPI_Bcast (members,nUniqueCount, MPI_INT,0,MPI_COMM_WORLD);		
					
		//Creating a new comm group
		MPI_Comm_group(MPI_COMM_WORLD, &MPI_GROUP_WORLD); 

		MPI_Group_incl(MPI_GROUP_WORLD, nUniqueCount, members, &MPI_GROUP_MT); 
		MPI_Comm_create(MPI_COMM_WORLD, MPI_GROUP_MT, &MPI_COMM_MT); 

		//Finding myself in members
		bool bActive = false;
		for (i=0; i<nSize; i++)
			if (members[i]==nRank) bActive=true;

		if (nRank!=0) delete [] members;

		if (bActive)
		{
			MPI_Group_rank (MPI_GROUP_MT, &g_nNode); 
			MPI_Group_size (MPI_GROUP_MT, &g_nSize); 
#ifdef DEBUG_MPI_STARTUP
			printf("Active process #%d has new rank #%d!\n",nRank, g_nNode);
#endif //DEBUG_MPI_STARTUP
		}
		else
		{
			//Waiting for die signal
			int die=0;
#ifdef _RECV_CPUBURST
			MPI_Status status;
			while (true)
			{
				int flag=0;
				MPI_Iprobe(0, TAG_DIE, MPI_COMM_WORLD, &flag, &status);

				if (flag) break;

				msleep2(2000);
			}
#else //_RECV_CPUBURST
			MPI_Recv (&die,1,MPI_INT,0,TAG_DIE,MPI_COMM_WORLD,&status);
#endif //_RECV_CPUBURST
			MPI_Barrier(MPI_COMM_WORLD);
#ifdef DEBUG_MPI_STARTUP
			printf("%d>> Passive rank passed last barrier.\n",nRank);
#endif //DEBUG_MPI_STARTUP
			MPI_Finalize_our();
#ifdef DEBUG_MPI_STARTUP
			printf("Passive process #%d was finished!\n",nRank);
#endif //DEBUG_MPI_STARTUP
			return 0;
		}
	}
#ifdef DEBUG_MPI_STARTUP
	printf("MPI Processes=%d, CPUs=%d, CurW=%dm, CurG=%d, Threads=%d\n", nSize, g_nSize, nRank, g_nNode, g_nThreadsPerNode);
#endif // DEBUG_MPI_STARTUP
#endif // CLUSTER_LAM_MPI

	int ret = mod2d_run(nFirstShot,nLastShot);

#ifdef CLUSTER_LAM_MPI
	if (nRank==0 && nodes!=NULL)
	{
		//	Sending die signal to non-active members
		int j;
		for (i=1; i<nSize; i++)
		{
			if (!nodes[i].bSelected)
			{
				int die=1;
				MPI_Send (&die,1, MPI_INT,i,TAG_DIE,MPI_COMM_WORLD);		
			}
		}
		if (nodes!=NULL) delete [] nodes;
	}
	MPI_Barrier(MPI_COMM_WORLD);
#ifdef DEBUG_MPI_STARTUP
	printf("%d>> Active rank passed last barrier.\n",nRank);
#endif //DEBUG_MPI_STARTUP
	MPI_Finalize_our();
#ifdef DEBUG_MPI_STARTUP
	printf("%d>> Active rank finished.\n",nRank);
#endif //DEBUG_MPI_STARTUP
#endif // CLUSTER_LAM_MPI
	return ret;
}
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

